#include <sys/time.h>
#include <CL/opencl.h>
#include <stdio.h>
#include <malloc.h>
#include <iostream>
#include <chrono>
#include <string.h>

//#define mem_use_buf
#define mem_alloc_host
//#define default_buffer


cl_context context; 			// OpenCl context
cl_command_queue commandQueue; 	// OpenCl command queue
cl_platform_id platform;		// OpenCl platform
cl_device_id device;			// OpenCl device
cl_uint deviceCount;			// Number of available devices
cl_program program;				// OpenCl program 
size_t kernel_size;				// Byte size of kernel code
cl_kernel kernel;				// Kernel variable
cl_event event; 				// event variable

cl_mem bufferA;					// input 
cl_mem bufferB;					// input
cl_mem bufferC;					// output

size_t local_size;				// Work-group size
size_t global_size;				// Global size

char* readKernelSource(const char *fileName);
std::string checkError(cl_int err);

long long getTime(){ // Time measured in microseconds
	struct timeval time;

	int err = gettimeofday(&time, NULL);

	return static_cast<long long>(time.tv_sec * 1000000 + time.tv_usec);
}

size_t roundUp(int localSize, int globalSize){
	int r = globalSize % localSize;
	if (r == 0){
		return globalSize;
	} else {
		return globalSize + localSize - r;
	}
}

int main(){
	cl_int err;


	// Get the AMD platform
	err = clGetPlatformIDs(1, &platform, NULL);
	if (err != CL_SUCCESS){
		std::cout << "Error in clGetPlatformIDs" << "\n";
		exit(-1);
	}

	// Get the AMD devices
	err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 0, NULL, &deviceCount);
	std::cout << "Number of devices: " << deviceCount << "\n";
	err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, deviceCount, &device, NULL);
	if (err != CL_SUCCESS){
		std::cout << "Error in clGetDeviceIDs" << "\n";
		exit(-1);
	}

	// Print device info
	char buffer[10240];
	cl_uint buf_uint;
	cl_ulong buf_ulong;
	err = clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(buffer), buffer, NULL);
	printf("  DEVICE_NAME = %s\n", buffer);
	err = clGetDeviceInfo(device, CL_DEVICE_VENDOR, sizeof(buffer), buffer, NULL);
	printf("  DEVICE_VENDOR = %s\n", buffer);
	err = clGetDeviceInfo(device, CL_DEVICE_VERSION, sizeof(buffer), buffer, NULL);
	printf("  DEVICE_VERSION = %s\n", buffer);
	err = clGetDeviceInfo(device, CL_DRIVER_VERSION, sizeof(buffer), buffer, NULL);
	printf("  DRIVER_VERSION = %s\n", buffer);
	err = clGetDeviceInfo(device, CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(buf_uint), &buf_uint, NULL);
	printf("  DEVICE_MAX_COMPUTE_UNITS = %u\n", (unsigned int)buf_uint);
	err = clGetDeviceInfo(device, CL_DEVICE_MAX_CLOCK_FREQUENCY, sizeof(buf_uint), &buf_uint, NULL);
	printf("  DEVICE_MAX_CLOCK_FREQUENCY = %u\n", (unsigned int)buf_uint);
	err = clGetDeviceInfo(device, CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(buf_ulong), &buf_ulong, NULL);
	printf("  DEVICE_GLOBAL_MEM_SIZE = %llu\n", (unsigned long long)buf_ulong);


	// Create the context
	context = clCreateContext(0,deviceCount, &device, NULL, NULL, &err);
	if (err != CL_SUCCESS){
		std::cout << "Error in clCreateContext" << "\n";
		exit(-1);
	}

	// Create command queue for GPU
	commandQueue = clCreateCommandQueue(context, device, CL_QUEUE_PROFILING_ENABLE, &err);
	if (err != CL_SUCCESS){
		std::cout << "Error in clCreateCommandQueue" << "\n";
		exit(-1);
	}

	// Read the OpenCL kernel from source file
	char* source = readKernelSource("test.cl");

	// Create a program with kernel source file
	program = clCreateProgramWithSource(context, 1, (const char **)&source, NULL, &err);
	if (err != CL_SUCCESS){
		std::cout << "Error in clCreateProgramWithSource" << "\n";
		exit(-1);
	}

	// Build the program
	err = clBuildProgram(program, deviceCount, &device, NULL, NULL, NULL);
	if (err != CL_SUCCESS){
		std::cout << "Error: " << checkError(err) << "\n";
		std::cout << "Error in clBuildProgram" << "\n";
		exit(-1);
	}

	// Create kernel
	kernel =clCreateKernel(program, "test", &err);
	if (err != CL_SUCCESS){
		std::cout << "Error in clCreateKernel: " << checkError(err) << "\n";
		exit(-1);
	}

// -------------- Start experiment ---------------------

	// Set local work-group size and global size
	int ELE_NUM = 10000000;
	local_size = 64;
	global_size = roundUp((int)local_size, ELE_NUM);
	


	// Allocate mem

	long long start_fulltime = getTime();

	float* A = (float*)malloc(sizeof(float)*ELE_NUM);
	float* B = (float*)malloc(sizeof(float)*ELE_NUM);
	float* C = (float*)malloc(sizeof(float)*ELE_NUM);

	int nloop = 100;
	long long start, end,pfill_s,pfill_e,pfill_time,a_st,a_en,b_st,b_en,c_st,c_en,a_time,b_time,c_time;

	long long creat_buffer = 0;
	long long q_sub = 0;
	long long sub_s = 0;
	long long s_e = 0;
	long long q_e = 0;
	long long release_buffer = 0;
	//long long start_fulltime = getTime();
	for (int i = 0; i < nloop; i++){
//		std::cout << " --- Loop # " << i << " --- \n";

		// Initialize two inputs and output
//		for (int i = 0; i < ELE_NUM; i++){
//			A[i] = 100.0;
//			B[i] = 11.1;
//			C[i] = 0.0;
//		}

		pfill_s=getTime();
		for (int i = 0; i < ELE_NUM; i++){
			A[i] = 100.0;
			B[i] = 11.1;

		}
		pfill_e=getTime();
		pfill_time=pfill_time+(pfill_e-pfill_s);

		// Create buffers
#ifdef mem_use_buf		
		start = getTime();
		bufferA = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR, sizeof(cl_float) * ELE_NUM, A, &err);
		bufferB = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR, sizeof(cl_float) * ELE_NUM, B, &err);
		//bufferC = clCreateBuffer(context, CL_MEM_WRITE_ONLY | CL_MEM_USE_HOST_PTR, sizeof(cl_float) * ELE_NUM, C, &err);
		
		end = getTime();
		creat_buffer += (end-start);
		bufferC = clCreateBuffer(context, CL_MEM_WRITE_ONLY , sizeof(cl_float) * ELE_NUM, NULL, &err);
	
#endif

#ifdef mem_alloc_host		

		start = getTime();
		a_st=getTime();
		bufferA = clCreateBuffer(context,  CL_MEM_ALLOC_HOST_PTR, sizeof(cl_float) * ELE_NUM, NULL, &err);
		cl_float* src_a=(cl_float*)clEnqueueMapBuffer(commandQueue, bufferA,CL_TRUE,CL_MAP_WRITE, 0, sizeof(cl_float) * ELE_NUM, 0, NULL, NULL, &err);

		bufferB = clCreateBuffer(context, CL_MEM_ALLOC_HOST_PTR, sizeof(cl_float) * ELE_NUM, NULL, &err);
		cl_float* src_b=(cl_float*)clEnqueueMapBuffer(commandQueue, bufferB,CL_TRUE,CL_MAP_WRITE, 0, sizeof(cl_float) * ELE_NUM, 0, NULL, NULL, &err);
		clFinish(commandQueue);
		a_en=getTime();
		a_time=a_time+(a_en-a_st);

		c_st=getTime();
		memcpy(src_a,A,sizeof(cl_float) * ELE_NUM);

		memcpy(src_b,B,sizeof(cl_float) * ELE_NUM);
		c_en=getTime();
		c_time=c_time+(c_en-c_st);

		b_st=getTime();
		clEnqueueUnmapMemObject(commandQueue, bufferB, src_b, 0, NULL, NULL);
		clEnqueueUnmapMemObject(commandQueue, bufferA, src_a, 0, NULL, NULL);
		clFinish(commandQueue);
		b_en=getTime();
		b_time=b_time+(b_en-b_st);


		end = getTime();
		creat_buffer += (end-start);
		bufferC = clCreateBuffer(context, CL_MEM_ALLOC_HOST_PTR, sizeof(cl_float) * ELE_NUM, NULL, &err);

#endif



//		std::cout << "clCreateBuffer: " << (end-start) << "\n";

		// Set kernel args
		err = clSetKernelArg(kernel, 0, sizeof(cl_mem), (void*)&bufferA);
		err = clSetKernelArg(kernel, 1, sizeof(cl_mem), (void*)&bufferB);
		err = clSetKernelArg(kernel, 2, sizeof(cl_mem), (void*)&bufferC);
		err = clSetKernelArg(kernel, 3, sizeof(int), (void*)&ELE_NUM);

		if (err != CL_SUCCESS){
			std::cout << "Error in clSetKernelArg" << "\n";
			exit(-1);
		}

		cl_ulong startTime;
		cl_ulong endTime;
		cl_ulong s,e,q,sub;
	
		// Kernel launch
		err = clEnqueueNDRangeKernel(commandQueue, kernel, 1, NULL, &global_size, &local_size, 0, NULL, &event);
		err = clFinish(commandQueue);

		if (clWaitForEvents(1 , &event) == CL_SUCCESS){ // wait for kernel execution

			err = clGetEventProfilingInfo(event,CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &s, NULL);

			err = clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &e, NULL);

			err = clGetEventProfilingInfo(event,CL_PROFILING_COMMAND_QUEUED, sizeof(cl_ulong), &q, NULL);

			err = clGetEventProfilingInfo(event,CL_PROFILING_COMMAND_SUBMIT, sizeof(cl_ulong), &sub, NULL);

			q_sub += static_cast<long long>(sub-q)/1000;
			sub_s += static_cast<long long>(s-sub)/1000;
			s_e += static_cast<long long>(e-s)/1000;
			q_e += static_cast<long long>(e-q)/1000;

//			printf("QUEUE -> SUBMIT: %lld\n", static_cast<long long>(sub-q)/1000);
//			printf("SUBMIT -> START: %lld\n", static_cast<long long>(s-sub)/1000);
//			printf("START -> END: %lld\n", static_cast<long long>(e-s)/1000);
//			printf("QUEUE -> END: %lld\n", static_cast<long long>(e-q)/1000);
		}

		
		//reading data from C
#ifndef mem_alloc_host	
		cl_int cerror;
		cerror=clEnqueueReadBuffer(commandQueue, bufferC,CL_TRUE, 0,sizeof(cl_float) * ELE_NUM,C, 0, NULL, NULL);
#endif
#ifdef mem_alloc_host	
	cl_int cerror;
	cl_float* output=(cl_float*)clEnqueueMapBuffer(commandQueue, bufferC,CL_TRUE,CL_MAP_READ, 0, sizeof(cl_float) * ELE_NUM, 0, NULL, NULL, &cerror);
	//memcpy(C,output,sizeof(cl_float) * ELE_NUM);
//	for (int i = 0; i < ELE_NUM; i++){
//				printf("c value is : %f\n",output[i]);
//			}
	clEnqueueUnmapMemObject(commandQueue, bufferC, output, 0, NULL, NULL);
#endif

		//
		
		// Release mem objects
		start = getTime();
		if(bufferA)clReleaseMemObject(bufferA);
		if(bufferB)clReleaseMemObject(bufferB);
		if(bufferC)clReleaseMemObject(bufferC);
		end = getTime();
//		std::cout << "clReleaseMemObject: " << (end-start) << "\n";
		release_buffer += (end-start);

//		// Verify outputs
//		float* test = new float[ELE_NUM];
//		for (int i = 0; i < ELE_NUM; i++){
//			test[i] = src_a[i] + src_b[i];
//		}
//
//		bool correct = true;
//		for (int i = 0; i < ELE_NUM; i++){
//
//		//	printf("output values: %f\n",C[i]);
//			if (test[i] != output[i]){
//				correct = false;
//			}
//		}
//
//		if (correct == true){
//			//printf("TRUE\n");
//			continue;
//		} else {
//			printf("FALSE\n");
//		}


	} // end of loop
	long long end_fulltime = getTime();

	std::cout << "clCreateBuffer: " << creat_buffer/nloop << "\n";
	std::cout << "QUEUE -> SUBMIT: " << q_sub/nloop << "\n";
	std::cout << "SUBMIT -> START: " << sub_s/nloop << "\n";
	std::cout << "START -> END: " << s_e/nloop << "\n";
	std::cout << "QUEUE -> END: " << q_e/nloop << "\n";
	std::cout << "clReleaseMemObject: " << release_buffer/nloop << "\n";
	std::cout << "pfill time: " << pfill_time/nloop << "\n";
	std::cout << "mapping time: " << a_time/nloop << "\n";
	std::cout << "unmapping time: " << b_time/nloop << "\n";
	std::cout << "memcpy time: " << c_time/nloop << "\n";

	std::cout << "Total loop execution time: " << (end_fulltime-start_fulltime) << std::endl;


  	if(commandQueue)clReleaseCommandQueue(commandQueue);
	if(context)clReleaseContext(context);
	if(program)clReleaseProgram(program);
	if(kernel)clReleaseKernel(kernel);
	//if(device)clReleaseDevice(device);

//	free(A);
//	free(B);
//	free(C);
	return 0;
}

char* readKernelSource(const char *fileName){
	/*
	 * Load the kernel source code into *char source
	 */
	FILE *fp;
	char *source;
	size_t size;

	fp = fopen(fileName, "r");
	if (!fp) {
		fprintf(stderr, "Failed to load kernel.\n");
		exit(1);
	}

	// Measure the size of source file
	fseek(fp, 0, SEEK_END);
	size = ftell(fp);

	fseek(fp, 0, SEEK_SET);

	source = (char*)malloc(size+1);
	size = fread(source, 1, size, fp);
	source[size]='\0';
	fclose( fp );

	return source;
}

std::string checkError(cl_int err){
	switch (err) {
		case CL_SUCCESS:                            return "Success!";
		case CL_DEVICE_NOT_FOUND:                   return "Device not found.";
		case CL_DEVICE_NOT_AVAILABLE:               return "Device not available";
		case CL_COMPILER_NOT_AVAILABLE:             return "Compiler not available";
		case CL_MEM_OBJECT_ALLOCATION_FAILURE:      return "Memory object allocation failure";
		case CL_OUT_OF_RESOURCES:                   return "Out of resources";
		case CL_OUT_OF_HOST_MEMORY:                 return "Out of host memory";
		case CL_PROFILING_INFO_NOT_AVAILABLE:       return "Profiling information not available";
		case CL_MEM_COPY_OVERLAP:                   return "Memory copy overlap";
		case CL_IMAGE_FORMAT_MISMATCH:              return "Image format mismatch";
		case CL_IMAGE_FORMAT_NOT_SUPPORTED:         return "Image format not supported";
		case CL_BUILD_PROGRAM_FAILURE:              return "Program build failure";
		case CL_MAP_FAILURE:                        return "Map failure";
		case CL_INVALID_VALUE:                      return "Invalid value";
		case CL_INVALID_DEVICE_TYPE:                return "Invalid device type";
		case CL_INVALID_PLATFORM:                   return "Invalid platform";
		case CL_INVALID_DEVICE:                     return "Invalid device";
		case CL_INVALID_CONTEXT:                    return "Invalid context";
		case CL_INVALID_QUEUE_PROPERTIES:           return "Invalid queue properties";
		case CL_INVALID_COMMAND_QUEUE:              return "Invalid command queue";
		case CL_INVALID_HOST_PTR:                   return "Invalid host pointer";
		case CL_INVALID_MEM_OBJECT:                 return "Invalid memory object";
		case CL_INVALID_IMAGE_FORMAT_DESCRIPTOR:    return "Invalid image format descriptor";
		case CL_INVALID_IMAGE_SIZE:                 return "Invalid image size";
		case CL_INVALID_SAMPLER:                    return "Invalid sampler";
		case CL_INVALID_BINARY:                     return "Invalid binary";
		case CL_INVALID_BUILD_OPTIONS:              return "Invalid build options";
		case CL_INVALID_PROGRAM:                    return "Invalid program";
		case CL_INVALID_PROGRAM_EXECUTABLE:         return "Invalid program executable";
		case CL_INVALID_KERNEL_NAME:                return "Invalid kernel name";
		case CL_INVALID_KERNEL_DEFINITION:          return "Invalid kernel definition";
		case CL_INVALID_KERNEL:                     return "Invalid kernel";
		case CL_INVALID_ARG_INDEX:                  return "Invalid argument index";
		case CL_INVALID_ARG_VALUE:                  return "Invalid argument value";
		case CL_INVALID_ARG_SIZE:                   return "Invalid argument size";
		case CL_INVALID_KERNEL_ARGS:                return "Invalid kernel arguments";
		case CL_INVALID_WORK_DIMENSION:             return "Invalid work dimension";
		case CL_INVALID_WORK_GROUP_SIZE:            return "Invalid work group size";
		case CL_INVALID_WORK_ITEM_SIZE:             return "Invalid work item size";
		case CL_INVALID_GLOBAL_OFFSET:              return "Invalid global offset";
		case CL_INVALID_EVENT_WAIT_LIST:            return "Invalid event wait list";
		case CL_INVALID_EVENT:                      return "Invalid event";
		case CL_INVALID_OPERATION:                  return "Invalid operation";
		case CL_INVALID_GL_OBJECT:                  return "Invalid OpenGL object";
		case CL_INVALID_BUFFER_SIZE:                return "Invalid buffer size";
		case CL_INVALID_MIP_LEVEL:                  return "Invalid mip-map level";
		default: return "Unknown";
	}
}
